@@ -1,4 +1,5 @@
-/* $Id: items.c,v 1.4 2003/06/14 02:27:53 bradfitz Exp $ */
+/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
+/* $Id: items.c,v 1.17 2003/06/27 20:32:30 bradfitz Exp $ */
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -18,7 +19,7 @@
 #include <event.h>
 #include <malloc.h>
 #include <Judy.h>
-
+#include <assert.h>
 
 #include "memcached.h"
 
@@ -55,7 +56,7 @@ item *item_alloc(char *key, int flags, time_t exptime, int nbytes) {
     if (id == 0)
         return 0;
 
-    it = slabs_alloc(id);
+    it = slabs_alloc(ntotal);
     if (it == 0) {
         /* 
          * try to get one off the right LRU 
@@ -76,35 +77,49 @@ item *item_alloc(char *key, int flags, time_t exptime, int nbytes) {
                 break;
             }
         }
-        it = slabs_alloc(id);
+        it = slabs_alloc(ntotal);
         if (it==0) return 0;
     }
 
+    assert(it->slabs_clsid == 0);
 
     it->slabs_clsid = id;
 
-    it->next = it->prev = 0;
+    assert(it != heads[it->slabs_clsid]);
+
+    it->next = it->prev = it->h_next = 0;
     it->refcount = 0;
     it->it_flags = 0;
-    it->key = (char *)&(it->end[0]);
-    it->data = (void *) (it->key + len);
-    strcpy(it->key, key);
-    it->exptime = exptime;
+    it->nkey = len;
     it->nbytes = nbytes;
-    it->ntotal = ntotal;
+    strcpy(ITEM_key(it), key);
+    it->exptime = exptime;
     it->flags = flags;
     return it;
 }
 
 void item_free(item *it) {
-    slabs_free(it, it->slabs_clsid);
+    unsigned int ntotal = ITEM_ntotal(it);
+    assert((it->it_flags & ITEM_LINKED) == 0);
+    assert(it != heads[it->slabs_clsid]);
+    assert(it != tails[it->slabs_clsid]);
+    assert(it->refcount == 0);    
+
+    /* so slab size changer can tell later if item is already free or not */
+    it->slabs_clsid = 0;
+    it->it_flags |= ITEM_SLABBED;
+    slabs_free(it, ntotal);
 }
 
 void item_link_q(item *it) { /* item is the new head */
     item **head, **tail;
-    if (it->slabs_clsid > LARGEST_ID) return;
+    assert(it->slabs_clsid <= LARGEST_ID);
+    assert((it->it_flags & ITEM_SLABBED) == 0);
+
     head = &heads[it->slabs_clsid];
     tail = &tails[it->slabs_clsid];
+    assert(it != *head);
+    assert((*head && *tail) || (*head == 0 && *tail == 0));
     it->prev = 0;
     it->next = *head;
     if (it->next) it->next->prev = it;
@@ -116,11 +131,21 @@ void item_link_q(item *it) { /* item is the new head */
 
 void item_unlink_q(item *it) {
     item **head, **tail;
-    if (it->slabs_clsid > LARGEST_ID) return;
+    assert(it->slabs_clsid <= LARGEST_ID);
     head = &heads[it->slabs_clsid];
     tail = &tails[it->slabs_clsid];
-    if (*head == it) *head = it->next;
-    if (*tail == it) *tail = it->prev;
+    
+    if (*head == it) {
+        assert(it->prev == 0);
+        *head = it->next;
+    }
+    if (*tail == it) {
+        assert(it->next == 0);
+        *tail = it->prev;
+    }
+    assert(it->next != it);
+    assert(it->prev != it);
+
     if (it->next) it->next->prev = it->prev;
     if (it->prev) it->prev->next = it->next;
     sizes[it->slabs_clsid]--;
@@ -128,11 +153,13 @@ void item_unlink_q(item *it) {
 }
 
 int item_link(item *it) {
+    assert((it->it_flags & (ITEM_LINKED|ITEM_SLABBED)) == 0);
+    assert(it->nbytes < 1048576);
     it->it_flags |= ITEM_LINKED;
     it->time = time(0);
-    assoc_insert(it->key, (void *)it);
+    assoc_insert(ITEM_key(it), it);
 
-    stats.curr_bytes += it->ntotal;
+    stats.curr_bytes += ITEM_ntotal(it);
     stats.curr_items += 1;
     stats.total_items += 1;
 
@@ -142,29 +169,36 @@ int item_link(item *it) {
 }
 
 void item_unlink(item *it) {
-    it->it_flags &= ~ITEM_LINKED;
-    assoc_delete(it->key);
-    item_unlink_q(it);
-    stats.curr_bytes -= it->ntotal;
-    stats.curr_items -= 1;
+    if (it->it_flags & ITEM_LINKED) {
+        it->it_flags &= ~ITEM_LINKED;
+        stats.curr_bytes -= ITEM_ntotal(it);
+        stats.curr_items -= 1;
+        assoc_delete(ITEM_key(it));
+        item_unlink_q(it);
+    }
     if (it->refcount == 0) item_free(it);
-    return;
 }
 
 void item_remove(item *it) {
+    assert((it->it_flags & ITEM_SLABBED) == 0);
     if (it->refcount) it->refcount--;
+    assert((it->it_flags & ITEM_DELETED) == 0 || it->refcount);
     if (it->refcount == 0 && (it->it_flags & ITEM_LINKED) == 0) {
         item_free(it);
     }
 }
 
 void item_update(item *it) {
+    assert((it->it_flags & ITEM_SLABBED) == 0);
+
     item_unlink_q(it);
     it->time = time(0);
     item_link_q(it);
 }
 
 int item_replace(item *it, item *new_it) {
+    assert((it->it_flags & ITEM_SLABBED) == 0);
+
     item_unlink(it);
     return item_link(new_it);
 }
@@ -191,7 +225,7 @@ char *item_cachedump(unsigned int slabs_clsid, unsigned int limit, unsigned int
             break;
         if (!it)
             break;
-        sprintf(temp, "ITEM %s [%u b; %lu s]\r\n", it->key, it->nbytes - 2, it->time);
+        sprintf(temp, "ITEM %s [%u b; %lu s]\r\n", ITEM_key(it), it->nbytes - 2, it->time);
         len = strlen(temp);
         if (bufcurr + len +5 > memlimit)  /* 5 is END\r\n */
             break;
@@ -245,8 +279,9 @@ char* item_stats_sizes(int *bytes) {
     for (i=0; i<LARGEST_ID; i++) {
         item *iter = heads[i];
         while (iter) {
-            int bucket = iter->ntotal / 32;
-            if (iter->ntotal % 32) bucket++;
+            int ntotal = ITEM_ntotal(iter);
+            int bucket = ntotal / 32;
+            if (ntotal % 32) bucket++;
             if (bucket < num_buckets) histogram[bucket]++;
             iter = iter->next;
         }
