@@ -13,7 +13,7 @@
  *      Anatoly Vorobey <mellon@pobox.com>
  *      Brad Fitzpatrick <brad@danga.com>
  *
- *  $Id: memcached.c,v 1.13 2003/06/15 23:27:18 bradfitz Exp $
+ *  $Id: memcached.c,v 1.26 2003/06/27 21:15:55 bradfitz Exp $
  */
 
 #include "config.h"
@@ -35,7 +35,7 @@
 #include <time.h>
 #include <event.h>
 #include <malloc.h>
-#include <Judy.h>
+#include <assert.h>
 
 #include "memcached.h"
 
@@ -46,36 +46,6 @@ static item **todelete = 0;
 static int delcurr;
 static int deltotal;
 
-/* associative array, using Judy */
-static Pvoid_t PJSLArray = (Pvoid_t) NULL;
-
-void assoc_init(void) {
-    return;
-}
-
-void *assoc_find(char *key) {
-    Word_t * PValue;  
-    JSLG( PValue, PJSLArray, key);
-    if (PValue) {
-        return ((void *)*PValue);
-    } else return 0;
-}
-
-int assoc_insert(char *key, void *value) {
-    Word_t *PValue;
-    JSLI( PValue, PJSLArray, key);
-    if (PValue) {
-        *PValue = (Word_t) value;
-        return 1;
-    } else return 0;
-}
-
-void assoc_delete(char *key) {
-    int Rc_int;
-    JSLD( Rc_int, PJSLArray, key);
-    return;
-}
-
 void stats_init(void) {
     stats.curr_items = stats.total_items = stats.curr_conns = stats.total_conns = stats.conn_structs = 0;
     stats.get_cmds = stats.set_cmds = stats.get_hits = stats.get_misses = 0;
@@ -245,12 +215,12 @@ void complete_nread(conn *c) {
     stats.set_cmds++;
 
     while(1) {
-        if (strncmp((char *)(it->data) + it->nbytes - 2, "\r\n", 2) != 0) {
+        if (strncmp(ITEM_data(it) + it->nbytes - 2, "\r\n", 2) != 0) {
             out_string(c, "CLIENT_ERROR bad data chunk");
             break;
         }
 
-        old_it = (item *)assoc_find(it->key);
+        old_it = assoc_find(ITEM_key(it));
 
         if (old_it && old_it->exptime && old_it->exptime < now) {
             item_unlink(old_it);
@@ -291,12 +261,13 @@ void process_stat(conn *c, char *command) {
     time_t now = time(0);
 
     if (strcmp(command, "stats") == 0) {
-        char temp[768];
+        char temp[1024];
         pid_t pid = getpid();
         char *pos = temp;
 
         pos += sprintf(pos, "STAT pid %u\r\n", pid);
         pos += sprintf(pos, "STAT uptime %lu\r\n", now - stats.started);
+        pos += sprintf(pos, "STAT time %u\r\n", now);
         pos += sprintf(pos, "STAT version " VERSION "\r\n");
         pos += sprintf(pos, "STAT curr_items %u\r\n", stats.curr_items);
         pos += sprintf(pos, "STAT total_items %u\r\n", stats.total_items);
@@ -310,7 +281,7 @@ void process_stat(conn *c, char *command) {
         pos += sprintf(pos, "STAT get_misses %u\r\n", stats.get_misses);
         pos += sprintf(pos, "STAT bytes_read %llu\r\n", stats.bytes_read);
         pos += sprintf(pos, "STAT bytes_written %llu\r\n", stats.bytes_written);
-        pos += sprintf(pos, "STAT limit_maxbytes %llu\r\n", settings.maxbytes);
+        pos += sprintf(pos, "STAT limit_maxbytes %u\r\n", settings.maxbytes);
         pos += sprintf(pos, "END");
         out_string(c, temp);
         return;
@@ -325,10 +296,19 @@ void process_stat(conn *c, char *command) {
     if (strcmp(command, "stats malloc") == 0) {
         char temp[512];
         struct mallinfo info;
+        char *pos = temp;
 
         info = mallinfo();
-        sprintf(temp, "STAT arena_size %d\r\nSTAT free_chunks %d\r\nSTAT fastbin_blocks %d\r\nSTAT mmaped_regions %d\r\nSTAT mmapped_space %d\r\nSTAT max_total_alloc %d\r\nSTAT fastbin_space %d\r\nSTAT total_alloc %d\r\nSTAT total_free %d\r\nSTAT releasable_space %d\r\nEND", 
-                info.arena, info.ordblks, info.smblks, info.hblks, info.hblkhd, info.usmblks, info.fsmblks, info.uordblks, info.fordblks, info.keepcost);
+        pos += sprintf(pos, "STAT arena_size %d\r\n", info.arena);
+        pos += sprintf(pos, "STAT free_chunks %d\r\n", info.ordblks);
+        pos += sprintf(pos, "STAT fastbin_blocks %d\r\n", info.smblks);
+        pos += sprintf(pos, "STAT mmapped_regions %d\r\n", info.hblks);
+        pos += sprintf(pos, "STAT mmapped_space %d\r\n", info.hblkhd);
+        pos += sprintf(pos, "STAT max_total_alloc %d\r\n", info.usmblks);
+        pos += sprintf(pos, "STAT fastbin_space %d\r\n", info.fsmblks);
+        pos += sprintf(pos, "STAT total_alloc %d\r\n", info.uordblks);
+        pos += sprintf(pos, "STAT total_free %d\r\n", info.fordblks);
+        pos += sprintf(pos, "STAT releasable_space %d\r\nEND", info.keepcost);
         out_string(c, temp);
         return;
     }
@@ -397,9 +377,17 @@ void process_stat(conn *c, char *command) {
     }
 
     if (strcmp(command, "stats slabs")==0) {
-        char buffer[4096];
-        slabs_stats(buffer, 4096);
-        out_string(c, buffer);
+        int bytes = 0;
+        char *buf = slabs_stats(&bytes);
+        if (!buf) {
+            out_string(c, "SERVER_ERROR out of memory");
+            return;
+        }
+        c->write_and_free = buf;
+        c->wcurr = buf;
+        c->wbytes = bytes;
+        c->state = conn_write;
+        c->write_and_go = conn_read;
         return;
     }
 
@@ -443,15 +431,14 @@ void process_command(conn *c, char *command) {
         (strncmp(command, "set ", 4) == 0 && (comm = NREAD_SET)) ||
         (strncmp(command, "replace ", 8) == 0 && (comm = NREAD_REPLACE))) {
 
-        char s_comm[10];
-        char key[256];
+        char key[251];
         int flags;
         time_t expire;
         int len, res;
         item *it;
 
-        res = sscanf(command, "%s %s %u %lu %d\n", s_comm, key, &flags, &expire, &len);
-        if (res!=5 || strlen(key)==0 ) {
+        res = sscanf(command, "%*s %250s %u %lu %d\n", key, &flags, &expire, &len);
+        if (res!=4 || strlen(key)==0 ) {
             out_string(c, "CLIENT_ERROR bad command line format");
             return;
         }
@@ -466,7 +453,7 @@ void process_command(conn *c, char *command) {
 
         c->item_comm = comm;
         c->item = it;
-        c->rcurr = it->data;
+        c->rcurr = ITEM_data(it);
         c->rlbytes = it->nbytes;
         c->state = conn_nread;
         return;
@@ -475,17 +462,16 @@ void process_command(conn *c, char *command) {
     if ((strncmp(command, "incr ", 5) == 0 && (incr = 1)) ||
         (strncmp(command, "decr ", 5) == 0)) {
         char temp[32];
-        char s_comm[10];
         unsigned int value;
         item *it;
         unsigned int delta;
-        char key[255];
+        char key[251];
         int res;
         char *ptr;
         time_t now = time(0);
 
-        res = sscanf(command, "%s %s %u\n", s_comm, key, &delta);
-        if (res!=3 || strlen(key)==0 ) {
+        res = sscanf(command, "%*s %250s %u\n", key, &delta);
+        if (res!=2 || strlen(key)==0 ) {
             out_string(c, "CLIENT_ERROR bad command line format");
             return;
         }
@@ -504,7 +490,7 @@ void process_command(conn *c, char *command) {
             return;
         }
 
-        ptr = it->data;
+        ptr = ITEM_data(it);
         while (*ptr && (*ptr<'0' && *ptr>'9')) ptr++;
         
         value = atoi(ptr);
@@ -520,17 +506,17 @@ void process_command(conn *c, char *command) {
         res = strlen(temp);
         if (res + 2 > it->nbytes) { /* need to realloc */
             item *new_it;
-            new_it = item_alloc(it->key, it->flags, it->exptime, res + 2 );
+            new_it = item_alloc(ITEM_key(it), it->flags, it->exptime, res + 2 );
             if (new_it == 0) {
                 out_string(c, "SERVER_ERROR out of memory");
                 return;
             }
-            memcpy(new_it->data, temp, res);
-            memcpy((char *)(new_it->data) + res, "\r\n", 2);
+            memcpy(ITEM_data(new_it), temp, res);
+            memcpy(ITEM_data(new_it) + res, "\r\n", 2);
             item_replace(it, new_it);
         } else { /* replace in-place */
-            memcpy(it->data, temp, res);
-            memset(it->data + res, ' ', it->nbytes-res-2);
+            memcpy(ITEM_data(it), temp, res);
+            memset(ITEM_data(it) + res, ' ', it->nbytes-res-2);
         }
         out_string(c, temp);
         return;
@@ -539,16 +525,16 @@ void process_command(conn *c, char *command) {
     if (strncmp(command, "get ", 4) == 0) {
 
         char *start = command + 4;
-        char key[256];
+        char key[251];
         int next;
         int i = 0;
         item *it;
         time_t now = time(0);
 
-        while(sscanf(start, " %s%n", key, &next) >= 1) {
+        while(sscanf(start, " %250s%n", key, &next) >= 1) {
             start+=next;
             stats.get_cmds++;
-            it = (item *)assoc_find(key);
+            it = assoc_find(key);
             if (it && (it->it_flags & ITEM_DELETED)) {
                 it = 0;
             }
@@ -583,25 +569,25 @@ void process_command(conn *c, char *command) {
     }
 
     if (strncmp(command, "delete ", 7) == 0) {
-        char key [256];
+        char key[251];
         char *start = command+7;
         item *it;
 
-        sscanf(start, " %s", key);
+        sscanf(start, " %250s", key);
         it = assoc_find(key);
         if (!it) {
             out_string(c, "NOT_FOUND");
             return;
-        } else {
-            it->refcount++;
-            /* use its expiration time as its deletion time now */
-            it->exptime = time(0) + 4;
-            it->it_flags |= ITEM_DELETED;
-            todelete[delcurr++] = it;
-            if (delcurr >= deltotal) {
-                deltotal *= 2;
-                todelete = realloc(todelete, sizeof(item *)*deltotal);
-            }
+        }
+
+        it->refcount++;
+        /* use its expiration time as its deletion time now */
+        it->exptime = time(0) + 4;
+        it->it_flags |= ITEM_DELETED;
+        todelete[delcurr++] = it;
+        if (delcurr >= deltotal) {
+            deltotal *= 2;
+            todelete = realloc(todelete, sizeof(item *)*deltotal);
         }
         out_string(c, "DELETED");
         return;
@@ -622,6 +608,28 @@ void process_command(conn *c, char *command) {
         return;
     }
 
+    if (strncmp(command, "slabs reassign ", 15) == 0) {
+        int src, dst;
+        char *start = command+15;
+        if (sscanf(start, "%u %u\r\n", &src, &dst) == 2) {
+            int rv = slabs_reassign(src, dst);
+            if (rv == 1) {
+                out_string(c, "DONE");
+                return;
+            }
+            if (rv == 0) {
+                out_string(c, "CANT");
+                return;
+            }
+            if (rv == -1) {
+                out_string(c, "BUSY");
+                return;
+            }
+        }
+        out_string(c, "CLIENT_ERROR bogus command");
+        return;
+    }
+    
     out_string(c, "ERROR");
     return;
 }
@@ -887,10 +895,10 @@ void drive_machine(conn *c) {
         case conn_mwrite:
             /* 
              * we're writing ibytes bytes from iptr. iptr alternates between
-             * ibuf, where we build a string "VALUE...", and it->data for the 
+             * ibuf, where we build a string "VALUE...", and ITEM_data(it) for the 
              * current item. When we finish a chunk, we choose the next one using 
              * ipart, which has the following semantics: 0 - start the loop, 1 - 
-             * we finished ibuf, go to current it->data; 2 - we finished it->data,
+             * we finished ibuf, go to current ITEM_data(it); 2 - we finished ITEM_data(it),
              * move to the next item and build its ibuf; 3 - we finished all items, 
              * write "END".
              */
@@ -924,24 +932,26 @@ void drive_machine(conn *c) {
                 switch (c->ipart) {
                 case 1:
                     it = *(c->icurr);
-                    c->iptr = it->data;
+                    assert((it->it_flags & ITEM_SLABBED) == 0);
+                    c->iptr = ITEM_data(it);
                     c->ibytes = it->nbytes;
                     c->ipart = 2;
                     break;
                 case 2:
                     it = *(c->icurr);
                     item_remove(it);
-                    if (c->ileft <= 1) {
+                    c->ileft--;
+                    if (c->ileft <= 0) {
                         c->ipart = 3;
                         break;
                     } else {
-                        c->ileft--;
                         c->icurr++;
                     }
                     /* FALL THROUGH */
                 case 0:
                     it = *(c->icurr);
-                    sprintf(c->ibuf, "VALUE %s %u %u\r\n", it->key, it->flags, it->nbytes - 2);
+                    assert((it->it_flags & ITEM_SLABBED) == 0);
+                    sprintf(c->ibuf, "VALUE %s %u %u\r\n", ITEM_key(it), it->flags, it->nbytes - 2);
                     c->iptr = c->ibuf;
                     c->ibytes = strlen(c->iptr);
                     c->ipart = 1;
@@ -1034,6 +1044,13 @@ int server_socket(int port) {
     return sfd;
 }
 
+/* invoke right before gdb is called, on assert */
+void pre_gdb () {
+    int i = 0;
+    if(l_socket) close(l_socket);
+    for (i=3; i<=500; i++) close(i); /* so lame */
+    kill(getpid(), SIGABRT);
+}
 
 struct event deleteevent;
 
@@ -1049,12 +1066,14 @@ void delete_handler(int fd, short which, void *arg) {
         int i, j=0;
         time_t now = time(0);
         for (i=0; i<delcurr; i++) {
-            if (todelete[i]->exptime < now) {
-                /* no longer mark it deleted. it's now expired, same as dead */
-                todelete[i]->it_flags &= ~ITEM_DELETED;
-                todelete[i]->refcount--;
+            item *it = todelete[i];
+            if (it->exptime < now) {
+                assert(it->refcount > 0);
+                it->it_flags &= ~ITEM_DELETED;
+                item_unlink(it);
+                item_remove(it);
             } else {
-                todelete[j++] = todelete[i];
+                todelete[j++] = it;
             }
         }
         delcurr = j;
@@ -1148,10 +1167,10 @@ void usage_license(void) {
     return;
 }
 
+int l_socket=0;
 
 int main (int argc, char **argv) {
     int c;
-    int l_socket;
     conn *l_conn;
     struct in_addr addr;
     int lock_memory = 0;
