@@ -1,4 +1,5 @@
-/* $Id: memcached.h,v 1.4 2003/06/15 23:27:18 bradfitz Exp $ */
+/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
+/* $Id: memcached.h,v 1.15 2003/06/27 21:15:55 bradfitz Exp $ */
 
 #define DATA_BUFFER_SIZE 2048
 
@@ -19,7 +20,7 @@ struct stats {
 };
 
 struct settings {
-    unsigned long long maxbytes;
+    unsigned int maxbytes;
     int maxconns;
     int port;
     struct in_addr interface;
@@ -32,22 +33,31 @@ extern struct settings settings;
 #define ITEM_LINKED 1
 #define ITEM_DELETED 2
 
+/* temp */
+#define ITEM_SLABBED 4
+
 typedef struct _stritem {
     struct _stritem *next;
     struct _stritem *prev;
-    int    refcount; 
-    int    it_flags;
-    char   *key;
-    void   *data;
-    unsigned int slabs_clsid;
+    struct _stritem *h_next;  /* hash chain next */
+    unsigned short  refcount; 
+    unsigned short  flags;
     int    nbytes;  /* size of data */
-    int    ntotal;  /* size of this struct + key + data */
-    int    flags;
     time_t time;    /* least recent access */
     time_t exptime; /* expire time */
+    unsigned char it_flags;     /* ITEM_* above */
+    unsigned char slabs_clsid;
+    unsigned char nkey;         /* key length, with terminating null and padding */
+    unsigned char dummy1;
     void * end[0];
 } item;
 
+#define ITEM_key(item) ((char*)&((item)->end[0]))
+
+/* warning: don't use these macros with a function, as it evals its arg twice */
+#define ITEM_data(item) ((char*) &((item)->end[0]) + (item)->nkey)
+#define ITEM_ntotal(item) (sizeof(struct _stritem) + (item)->nkey + (item)->nbytes)
+
 enum conn_states {
     conn_listening,  /* the socket which listens for connections */
     conn_read,       /* reading in a command line */
@@ -88,7 +98,7 @@ typedef struct {
     /* 
      * item is used to hold an item structure created after reading the command
      * line of set/add/replace commands, but before we finished reading the actual 
-     * data. The data is read into item->data to avoid extra copying.
+     * data. The data is read into ITEM_data(item) to avoid extra copying.
      */
 
     void   *item;     /* for commands set/add/replace  */
@@ -109,6 +119,12 @@ typedef struct {
                          
 } conn;
 
+/* listening socket */
+extern int l_socket;
+
+/* temporary hack */
+/* #define assert(x) if(!(x)) { printf("assert failure: %s\n", #x); pre_gdb(); }
+   void pre_gdb (); */
 
 /*
  * Functions
@@ -123,15 +139,20 @@ void slabs_init(unsigned int limit);
 /* 0 means error: can't store such a large object */
 unsigned int slabs_clsid(unsigned int size);
 
-/* Allocate object of given size class identified by id. 0 on error */
-void *slabs_alloc(unsigned int id);
+/* Allocate object of given length. 0 on error */
+void *slabs_alloc(unsigned int size);
 
 /* Free previously allocated object */
-void slabs_free(void *ptr, unsigned int id);
+void slabs_free(void *ptr, unsigned int size);
     
 /* Fill buffer with stats */
-void slabs_stats(char *buffer, int buflen);
-        
+char* slabs_stats(int *buflen);
+
+/* Request some slab be moved between classes
+  1 = success
+   0 = fail
+   -1 = tried. busy. send again shortly. */
+int slabs_reassign(unsigned char srcid, unsigned char dstid);
     
 /* event handling, network IO */
 void event_handler(int fd, short which, void *arg);
@@ -156,8 +177,8 @@ void settings_init(void);
 
 /* associative array */
 void assoc_init(void);
-void *assoc_find(char *key);
-int assoc_insert(char *key, void *value);
+item *assoc_find(char *key);
+int assoc_insert(char *key, item *item);
 void assoc_delete(char *key);
 
 
